@@ -15,6 +15,7 @@
 #include <fcntl.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
 
 #include <glib.h>
 
@@ -46,7 +47,8 @@ struct SoupSSLCredentials {
 typedef struct {
 	GIOChannel channel;
 	GIOChannel *real_sock;
-	gboolean non_blocking;
+	int sockfd;
+	gboolean non_blocking, eagain;
 	gnutls_session session;
 	SoupSSLCredentials *creds;
 	char *hostname;
@@ -110,6 +112,7 @@ verify_certificate (gnutls_session session, const char *hostname, GError **err)
 			session, &cert_list_size);
 
 		if (cert_list == NULL) {
+			gnutls_x509_crt_deinit (cert);
 			g_set_error (err, SOUP_SSL_ERROR,
 				     SOUP_SSL_ERROR_CERTIFICATE,
 				     "No SSL certificate was found.");
@@ -118,6 +121,7 @@ verify_certificate (gnutls_session session, const char *hostname, GError **err)
 
 		if (gnutls_x509_crt_import (cert, &cert_list[0],
 					    GNUTLS_X509_FMT_DER) < 0) {
+			gnutls_x509_crt_deinit (cert);
 			g_set_error (err, SOUP_SSL_ERROR,
 				     SOUP_SSL_ERROR_CERTIFICATE,
 				     "The SSL certificate could not be parsed.");
@@ -125,11 +129,14 @@ verify_certificate (gnutls_session session, const char *hostname, GError **err)
 		}
 
 		if (!gnutls_x509_crt_check_hostname (cert, hostname)) {
+			gnutls_x509_crt_deinit (cert);
 			g_set_error (err, SOUP_SSL_ERROR,
 				     SOUP_SSL_ERROR_CERTIFICATE,
 				     "The SSL certificate does not match the hostname.");
 			return FALSE;
 		}
+
+		gnutls_x509_crt_deinit (cert);
 	}
 
 	return TRUE;
@@ -156,12 +163,15 @@ again:
 	}
 
 	if (result < 0) {
-		g_set_error (err, G_IO_CHANNEL_ERROR,
-			     G_IO_CHANNEL_ERROR_FAILED,
-			     "Unable to handshake");
+		g_set_error (err, SOUP_SSL_ERROR,
+			     SOUP_SSL_ERROR_HANDSHAKE_FAILED,
+			     "SSL handshake failed: %s",
+			     gnutls_strerror (result));
 		return G_IO_STATUS_ERROR;
 	}
 
+	chan->established = TRUE;
+
 	if (chan->type == SOUP_SSL_TYPE_CLIENT && chan->creds->have_ca_file &&
 	    !verify_certificate (chan->session, chan->hostname, err))
 		return G_IO_STATUS_ERROR;
@@ -188,8 +198,6 @@ again:
 		if (result == G_IO_STATUS_AGAIN ||
 		    result == G_IO_STATUS_ERROR)
 			return result;
-
-		chan->established = TRUE;
 	}
 
 	result = gnutls_record_recv (chan->session, buf, count);
@@ -200,12 +208,25 @@ again:
 	}
 
 	if (result == GNUTLS_E_INTERRUPTED || result == GNUTLS_E_AGAIN) {
-		if (chan->non_blocking)
+		if (chan->non_blocking || chan->eagain)
 			return G_IO_STATUS_AGAIN;
 		else
 			goto again;
 	}
 
+	if (result == GNUTLS_E_UNEXPECTED_PACKET_LENGTH) {
+		/* This means the connection was either corrupted or
+		 * interrupted. One particular thing that it can mean
+		 * is that the remote end closed the connection
+		 * abruptly without doing a proper TLS Close. There
+		 * are security reasons why it's bad to treat this as
+		 * not-an-error, but for compatibility reasons (eg,
+		 * bug 577386) we kinda have to. And it's not like
+		 * we're very secure anyway.
+		 */
+		return G_IO_STATUS_EOF;
+	}
+
 	if (result < 0) {
 		g_set_error (err, G_IO_CHANNEL_ERROR,
 			     G_IO_CHANNEL_ERROR_FAILED,
@@ -237,8 +258,6 @@ again:
 		if (result == G_IO_STATUS_AGAIN ||
 		    result == G_IO_STATUS_ERROR)
 			return result;
-
-		chan->established = TRUE;
 	}
 
 	result = gnutls_record_send (chan->session, buf, count);
@@ -252,7 +271,7 @@ again:
 	}
 
 	if (result == GNUTLS_E_INTERRUPTED || result == GNUTLS_E_AGAIN) {
-		if (chan->non_blocking)
+		if (chan->non_blocking || chan->eagain)
 			return G_IO_STATUS_AGAIN;
 		else
 			goto again;
@@ -368,6 +387,56 @@ init_dh_params (void)
 	return dh_params != NULL;
 }
 
+static ssize_t
+soup_gnutls_pull_func (gnutls_transport_ptr_t transport_data,
+		       void *buf, size_t buflen)
+{
+	SoupGNUTLSChannel *chan = transport_data;
+	ssize_t nread;
+
+	nread = recv (chan->sockfd, buf, buflen, 0);
+#ifdef G_OS_WIN32
+	{
+		int wsa_errno = WSAGetLastError ();
+		chan->eagain = (nread == SOCKET_ERROR && (wsa_errno == WSAEWOULDBLOCK ||
+							  wsa_errno == WSAETIMEDOUT));
+		if (nread == SOCKET_ERROR)
+			gnutls_transport_set_errno (chan->session,
+						    ((wsa_errno == WSAEWOULDBLOCK ||
+						      wsa_errno == WSAETIMEDOUT) ? EAGAIN :
+						     (wsa_errno == WSAEINTR ? EINTR :
+						      EIO)));
+	}
+#else
+	chan->eagain = (nread == -1 && errno == EAGAIN);
+#endif
+	return nread;
+}
+
+static ssize_t
+soup_gnutls_push_func (gnutls_transport_ptr_t transport_data,
+		       const void *buf, size_t buflen)
+{
+	SoupGNUTLSChannel *chan = transport_data;
+	ssize_t nwrote;
+
+	nwrote = send (chan->sockfd, buf, buflen, 0);
+#ifdef G_OS_WIN32
+	{
+		int wsa_errno = WSAGetLastError ();
+		chan->eagain = (nwrote == SOCKET_ERROR && wsa_errno == WSAEWOULDBLOCK);
+		if (nwrote == SOCKET_ERROR)
+			gnutls_transport_set_errno (chan->session,
+						    (wsa_errno == WSAEWOULDBLOCK ? EAGAIN :
+						     (wsa_errno == WSAEINTR ? EINTR :
+						      EIO)));
+	}
+#else
+	chan->eagain = (nwrote == -1 && errno == EAGAIN);
+#endif
+	return nwrote;
+}
+
 /**
  * soup_ssl_wrap_iochannel:
  * @sock: a #GIOChannel wrapping a TCP socket.
@@ -407,7 +476,8 @@ soup_ssl_wrap_iochannel (GIOChannel *sock, gboolean non_blocking,
 	if (ret)
 		goto THROW_CREATE_ERROR;
 
-	if (gnutls_set_default_priority (session) != 0)
+	/* See http://bugzilla.gnome.org/show_bug.cgi?id=581342 */
+	if (gnutls_priority_set_direct (session, "NORMAL:!VERS-TLS1.1:!VERS-TLS1.0", NULL) != 0)
 		goto THROW_CREATE_ERROR;
 
 	if (gnutls_credentials_set (session, GNUTLS_CRD_CERTIFICATE,
@@ -416,11 +486,15 @@ soup_ssl_wrap_iochannel (GIOChannel *sock, gboolean non_blocking,
 
 	if (type == SOUP_SSL_TYPE_SERVER)
 		gnutls_dh_set_prime_bits (session, DH_BITS);
-
-	gnutls_transport_set_ptr (session, GINT_TO_POINTER (sockfd));
+	else {
+		// gnutls defaults to requiring at least 768-bit keys,
+		// but there are some lame servers out there...
+		gnutls_dh_set_prime_bits (session, 256);
+	}
 
 	chan = g_slice_new0 (SoupGNUTLSChannel);
 	chan->real_sock = sock;
+	chan->sockfd = sockfd;
 	chan->session = session;
 	chan->creds = creds;
 	chan->hostname = g_strdup (remote_host);
@@ -428,6 +502,10 @@ soup_ssl_wrap_iochannel (GIOChannel *sock, gboolean non_blocking,
 	chan->non_blocking = non_blocking;
 	g_io_channel_ref (sock);
 
+	gnutls_transport_set_ptr (session, chan);
+	gnutls_transport_set_push_function (session, soup_gnutls_push_func);
+	gnutls_transport_set_pull_function (session, soup_gnutls_pull_func);
+
 	gchan = (GIOChannel *) chan;
 	gchan->funcs = (GIOFuncs *)&soup_gnutls_channel_funcs;
 	g_io_channel_init (gchan);
@@ -446,6 +524,54 @@ soup_ssl_wrap_iochannel (GIOChannel *sock, gboolean non_blocking,
 GCRY_THREAD_OPTION_PTHREAD_IMPL;
 #endif
 
+#ifdef G_OS_WIN32
+
+static int
+soup_gcry_win32_mutex_init (void **priv)
+{
+	int err = 0;
+	CRITICAL_SECTION *lock = (CRITICAL_SECTION*)malloc (sizeof (CRITICAL_SECTION));
+
+	if (!lock)
+		err = ENOMEM;
+	if (!err) {
+		InitializeCriticalSection (lock);
+		*priv = lock;
+	}
+	return err;
+}
+
+static int
+soup_gcry_win32_mutex_destroy (void **lock)
+{
+	DeleteCriticalSection ((CRITICAL_SECTION*)*lock);
+	free (*lock);
+	return 0;
+}
+
+static int
+soup_gcry_win32_mutex_lock (void **lock)
+{
+	EnterCriticalSection ((CRITICAL_SECTION*)*lock);
+	return 0;
+}
+
+static int
+soup_gcry_win32_mutex_unlock (void **lock)
+{
+	LeaveCriticalSection ((CRITICAL_SECTION*)*lock);
+	return 0;
+}
+
+
+static struct gcry_thread_cbs soup_gcry_threads_win32 = {		 \
+	(GCRY_THREAD_OPTION_USER | (GCRY_THREAD_OPTION_VERSION << 8)),	 \
+	NULL, soup_gcry_win32_mutex_init, soup_gcry_win32_mutex_destroy, \
+	soup_gcry_win32_mutex_lock, soup_gcry_win32_mutex_unlock,	 \
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
+
+#endif
+
 static void
 soup_gnutls_init (void)
 {
@@ -454,6 +580,8 @@ soup_gnutls_init (void)
 	if (g_once_init_enter (&inited_gnutls)) {
 #if defined(GCRY_THREAD_OPTION_PTHREAD_IMPL) && !defined(G_OS_WIN32)
 		gcry_control (GCRYCTL_SET_THREAD_CBS, &gcry_threads_pthread);
+#elif defined(G_OS_WIN32)
+		gcry_control (GCRYCTL_SET_THREAD_CBS, &soup_gcry_threads_win32);
 #endif
 		gnutls_global_init ();
 		g_once_init_leave (&inited_gnutls, TRUE);
@@ -486,6 +614,10 @@ soup_ssl_get_client_credentials (const char *ca_file)
 	creds = g_slice_new0 (SoupSSLCredentials);
 	gnutls_certificate_allocate_credentials (&creds->creds);
 
+	/* http://bugzilla.gnome.org/show_bug.cgi?id=589323 */
+	gnutls_certificate_set_verify_flags (creds->creds,
+					     GNUTLS_VERIFY_ALLOW_X509_V1_CA_CRT);
+
 	if (ca_file) {
 		creds->have_ca_file = TRUE;
 		status = gnutls_certificate_set_x509_trust_file (
