@@ -17,6 +17,7 @@
 #include "soup-message.h"
 #include "soup-session.h"
 #include "soup-session-feature.h"
+#include "soup-socket.h"
 #include "soup-uri.h"
 
 /**
@@ -67,16 +68,17 @@
  * used to identify when multiple messages are sent across the same
  * connection.
  *
- * The request half of the message is logged just before the first
- * byte of the request gets written to the network (from the
- * #SoupSession::request_started signal), and the response is logged
- * just after the last byte of the response body is read from the
- * network (from the #SoupMessage::got_body or
- * #SoupMessage::got_informational signal). In particular, note that
- * the #SoupMessage::got_headers signal, and anything triggered off it
- * (such as #SoupSession::authenticate) will be emitted
- * <emphasis>before</emphasis> the response headers are actually
- * logged.
+ * Currently, the request half of the message is logged just before
+ * the first byte of the request gets written to the network (from the
+ * #SoupSession::request_started signal), which means that if you have
+ * not made the complete request body available at that point, it will
+ * not be logged. The response is logged just after the last byte of
+ * the response body is read from the network (from the
+ * #SoupMessage::got_body or #SoupMessage::got_informational signal),
+ * which means that the #SoupMessage::got_headers signal, and anything
+ * triggered off it (such as #SoupSession::authenticate) will be
+ * emitted <emphasis>before</emphasis> the response headers are
+ * actually logged.
  **/
 
 static void soup_logger_session_feature_init (SoupSessionFeatureInterface *feature_interface, gpointer interface_data);
@@ -487,7 +489,7 @@ print_request (SoupLogger *logger, SoupMessage *msg,
 	if (log_level == SOUP_LOGGER_LOG_MINIMAL)
 		return;
 
-	if (!soup_message_headers_get (msg->request_headers, "Host")) {
+	if (!soup_message_headers_get_one (msg->request_headers, "Host")) {
 		soup_logger_print (logger, SOUP_LOGGER_LOG_HEADERS, '>',
 				   "Host: %s%c%u", uri->host,
 				   soup_uri_uses_default_port (uri) ? '\0' : ':',
@@ -506,10 +508,12 @@ print_request (SoupLogger *logger, SoupMessage *msg,
 	if (log_level == SOUP_LOGGER_LOG_HEADERS)
 		return;
 
-	if (msg->request_body->length) {
+	if (msg->request_body->length &&
+	    soup_message_body_get_accumulate (msg->request_body)) {
 		SoupBuffer *request;
 
 		request = soup_message_body_flatten (msg->request_body);
+		g_return_if_fail (request != NULL);
 		soup_buffer_free (request);
 
 		if (soup_message_headers_get_expectations (msg->request_headers) != SOUP_EXPECTATION_CONTINUE) {
@@ -560,7 +564,7 @@ print_response (SoupLogger *logger, SoupMessage *msg)
 	if (log_level == SOUP_LOGGER_LOG_HEADERS)
 		return;
 
-	if (msg->response_body->length) {
+	if (msg->response_body->data) {
 		soup_logger_print (logger, SOUP_LOGGER_LOG_BODY, '<',
 				   "\n%s", msg->response_body->data);
 	}
@@ -618,6 +622,8 @@ static void
 request_queued (SoupSessionFeature *logger, SoupSession *session,
 		SoupMessage *msg)
 {
+	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+
 	g_signal_connect (msg, "got-informational",
 			  G_CALLBACK (got_informational),
 			  logger);
@@ -634,6 +640,10 @@ request_started (SoupSessionFeature *feature, SoupSession *session,
 	gboolean restarted;
 	guint msg_id;
 
+	g_return_if_fail (SOUP_IS_SESSION (session));
+	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+	g_return_if_fail (SOUP_IS_SOCKET (socket));
+
 	msg_id = soup_logger_get_id (logger, msg);
 	if (msg_id)
 		restarted = TRUE;
@@ -656,6 +666,8 @@ static void
 request_unqueued (SoupSessionFeature *logger, SoupSession *session,
 		  SoupMessage *msg)
 {
+	g_return_if_fail (SOUP_IS_MESSAGE (msg));
+
 	g_signal_handlers_disconnect_by_func (msg, got_informational, logger);
 	g_signal_handlers_disconnect_by_func (msg, got_body, logger);
 }
