@@ -40,10 +40,15 @@
  * same handler, just pass "/" (or %NULL) for the path.) Any request
  * that does not match any handler will automatically be returned to
  * the client with a 404 (Not Found) status.
+ *
+ * If you want to handle the special "*" URI (eg, "OPTIONS *"), you
+ * must explicitly register a handler for "*"; the default handler
+ * will not be used for that case.
  * 
  * To add authentication to some or all paths, create an appropriate
  * #SoupAuthDomain (qv), and add it to the server via
- * soup_server_add_auth_domain.
+ * soup_server_add_auth_domain(). (As with handlers, you must
+ * explicitly add "*" to an auth domain if you want it to be covered.)
  * 
  * Additional processing options are available via #SoupServer's
  * signals; Connect to #SoupServer::request-started to be notified
@@ -76,6 +81,8 @@ struct SoupClientContext {
 	SoupSocket     *sock;
 	SoupAuthDomain *auth_domain;
 	char           *auth_user;
+
+	int             ref_count;
 };
 
 typedef struct {
@@ -93,6 +100,8 @@ typedef struct {
 	char              *ssl_cert_file, *ssl_key_file;
 	SoupSSLCredentials *ssl_creds;
 
+	char              *server_header;
+
 	GMainLoop         *loop;
 
 	SoupSocket        *listen_sock;
@@ -108,6 +117,8 @@ typedef struct {
 } SoupServerPrivate;
 #define SOUP_SERVER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_SERVER, SoupServerPrivate))
 
+#define SOUP_SERVER_SERVER_HEADER_BASE "libsoup/" PACKAGE_VERSION
+
 enum {
 	PROP_0,
 
@@ -117,6 +128,7 @@ enum {
 	PROP_SSL_KEY_FILE,
 	PROP_ASYNC_CONTEXT,
 	PROP_RAW_PATHS,
+	PROP_SERVER_HEADER,
 
 	LAST_PROP
 };
@@ -159,6 +171,8 @@ finalize (GObject *object)
 	if (priv->ssl_creds)
 		soup_ssl_free_server_credentials (priv->ssl_creds);
 
+	g_free (priv->server_header);
+
 	if (priv->listen_sock)
 		g_object_unref (priv->listen_sock);
 
@@ -305,6 +319,12 @@ soup_server_class_init (SoupServerClass *server_class)
 			      SOUP_TYPE_CLIENT_CONTEXT);
 
 	/* properties */
+	/**
+	 * SOUP_SERVER_PORT:
+	 *
+	 * Alias for the #SoupServer:port property. (The port the
+	 * server listens on.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_PORT,
 		g_param_spec_uint (SOUP_SERVER_PORT,
@@ -312,6 +332,12 @@ soup_server_class_init (SoupServerClass *server_class)
 				   "Port to listen on",
 				   0, 65536, 0,
 				   G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	/**
+	 * SOUP_SERVER_INTERFACE:
+	 *
+	 * Alias for the #SoupServer:interface property. (The address
+	 * of the network interface the server listens on.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_INTERFACE,
 		g_param_spec_object (SOUP_SERVER_INTERFACE,
@@ -319,6 +345,12 @@ soup_server_class_init (SoupServerClass *server_class)
 				     "Address of interface to listen on",
 				     SOUP_TYPE_ADDRESS,
 				     G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	/**
+	 * SOUP_SERVER_SSL_CERT_FILE:
+	 *
+	 * Alias for the #SoupServer:ssl-cert-file property. (The file
+	 * containing the SSL certificate for the server.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_SSL_CERT_FILE,
 		g_param_spec_string (SOUP_SERVER_SSL_CERT_FILE,
@@ -326,6 +358,12 @@ soup_server_class_init (SoupServerClass *server_class)
 				     "File containing server SSL certificate",
 				     NULL,
 				     G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	/**
+	 * SOUP_SERVER_SSL_KEY_FILE:
+	 *
+	 * Alias for the #SoupServer:ssl-key-file property. (The file
+	 * containing the SSL certificate key for the server.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_SSL_KEY_FILE,
 		g_param_spec_string (SOUP_SERVER_SSL_KEY_FILE,
@@ -333,12 +371,25 @@ soup_server_class_init (SoupServerClass *server_class)
 				     "File containing server SSL key",
 				     NULL,
 				     G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	/**
+	 * SOUP_SERVER_ASYNC_CONTEXT:
+	 *
+	 * Alias for the #SoupServer:async-context property. (The
+	 * server's #GMainContext.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_ASYNC_CONTEXT,
 		g_param_spec_pointer (SOUP_SERVER_ASYNC_CONTEXT,
 				      "Async GMainContext",
 				      "The GMainContext to dispatch async I/O in",
 				      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	/**
+	 * SOUP_SERVER_RAW_PATHS:
+	 *
+	 * Alias for the #SoupServer:raw-paths property. (If %TRUE,
+	 * percent-encoding in the Request-URI path will not be
+	 * automatically decoded.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_RAW_PATHS,
 		g_param_spec_boolean (SOUP_SERVER_RAW_PATHS,
@@ -346,6 +397,47 @@ soup_server_class_init (SoupServerClass *server_class)
 				      "If %TRUE, percent-encoding in the Request-URI path will not be automatically decoded.",
 				      FALSE,
 				      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+
+	/**
+	 * SoupServer:server-header:
+	 *
+	 * If non-%NULL, the value to use for the "Server" header on
+	 * #SoupMessage<!-- -->s processed by this server.
+	 *
+	 * The Server header is the server equivalent of the
+	 * User-Agent header, and provides information about the
+	 * server and its components. It contains a list of one or
+	 * more product tokens, separated by whitespace, with the most
+	 * significant product token coming first. The tokens must be
+	 * brief, ASCII, and mostly alphanumeric (although "-", "_",
+	 * and "." are also allowed), and may optionally include a "/"
+	 * followed by a version string. You may also put comments,
+	 * enclosed in parentheses, between or after the tokens.
+	 *
+	 * Some HTTP server implementations intentionally do not use
+	 * version numbers in their Server header, so that
+	 * installations running older versions of the server don't
+	 * end up advertising their vulnerability to specific security
+	 * holes.
+	 *
+	 * As with #SoupSession:user_agent, if you set a
+	 * %server_header property that has trailing whitespace,
+	 * #SoupServer will append its own product token (eg,
+	 * "<literal>libsoup/2.3.2</literal>") to the end of the
+	 * header for you.
+	 **/
+	/**
+	 * SOUP_SERVER_SERVER_HEADER:
+	 *
+	 * Alias for the #SoupServer:server-header property, qv.
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_SERVER_HEADER,
+		g_param_spec_string (SOUP_SERVER_SERVER_HEADER,
+				     "Server header",
+				     "Server header",
+				     NULL,
+				     G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
 }
 
 static GObject *
@@ -404,6 +496,7 @@ set_property (GObject *object, guint prop_id,
 	      const GValue *value, GParamSpec *pspec)
 {
 	SoupServerPrivate *priv = SOUP_SERVER_GET_PRIVATE (object);
+	const char *header;
 
 	switch (prop_id) {
 	case PROP_PORT:
@@ -432,7 +525,23 @@ set_property (GObject *object, guint prop_id,
 	case PROP_RAW_PATHS:
 		priv->raw_paths = g_value_get_boolean (value);
 		break;
+	case PROP_SERVER_HEADER:
+		g_free (priv->server_header);
+		header = g_value_get_string (value);
+		if (!header)
+			priv->server_header = NULL;
+		else if (!*header) {
+			priv->server_header =
+				g_strdup (SOUP_SERVER_SERVER_HEADER_BASE);
+		} else if (g_str_has_suffix (header, " ")) {
+			priv->server_header =
+				g_strdup_printf ("%s%s", header,
+						 SOUP_SERVER_SERVER_HEADER_BASE);
+		} else
+			priv->server_header = g_strdup (header);
+		break;
 	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
 	}
 }
@@ -462,7 +571,11 @@ get_property (GObject *object, guint prop_id,
 	case PROP_RAW_PATHS:
 		g_value_set_boolean (value, priv->raw_paths);
 		break;
+	case PROP_SERVER_HEADER:
+		g_value_set_string (value, priv->server_header);
+		break;
 	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
 	}
 }
@@ -473,6 +586,8 @@ get_property (GObject *object, guint prop_id,
  * @...: value of @optname1, followed by additional property/value pairs
  *
  * Creates a new #SoupServer.
+ *
+ * Return value: a new #SoupServer
  **/
 SoupServer *
 soup_server_new (const char *optname1, ...)
@@ -552,6 +667,18 @@ soup_server_get_listener (SoupServer *server)
 
 static void start_request (SoupServer *, SoupClientContext *);
 
+static SoupClientContext *
+soup_client_context_new (SoupServer *server, SoupSocket *sock)
+{
+	SoupClientContext *client = g_slice_new0 (SoupClientContext);
+
+	client->server = server;
+	client->sock = sock;
+	client->ref_count = 1;
+
+	return client;
+}
+
 static void
 soup_client_context_cleanup (SoupClientContext *client)
 {
@@ -565,6 +692,22 @@ soup_client_context_cleanup (SoupClientContext *client)
 	}
 }
 
+static SoupClientContext *
+soup_client_context_ref (SoupClientContext *client)
+{
+	client->ref_count++;
+	return client;
+}
+
+static void
+soup_client_context_unref (SoupClientContext *client)
+{
+	if (--client->ref_count == 0) {
+		soup_client_context_cleanup (client);
+		g_slice_free (SoupClientContext, client);
+	}
+}
+
 static void
 request_finished (SoupMessage *msg, SoupClientContext *client)
 {
@@ -582,7 +725,7 @@ request_finished (SoupMessage *msg, SoupClientContext *client)
 		start_request (server, client);
 	} else {
 		soup_socket_disconnect (sock);
-		g_slice_free (SoupClientContext, client);
+		soup_client_context_unref (client);
 	}
 	g_object_unref (msg);
 	g_object_unref (sock);
@@ -601,6 +744,8 @@ soup_server_get_handler (SoupServer *server, const char *path)
 		hand = soup_path_map_lookup (priv->handlers, path);
 		if (hand)
 			return hand;
+		if (!strcmp (path, "*"))
+			return NULL;
 	}
 	return priv->default_handler;
 }
@@ -674,6 +819,8 @@ call_handler (SoupMessage *req, SoupClientContext *client)
 	SoupServerHandler *hand;
 	SoupURI *uri;
 
+	g_signal_emit (server, signals[REQUEST_READ], 0, req, client);
+
 	if (req->status_code != 0)
 		return;
 
@@ -705,14 +852,19 @@ call_handler (SoupMessage *req, SoupClientContext *client)
 static void
 start_request (SoupServer *server, SoupClientContext *client)
 {
+	SoupServerPrivate *priv = SOUP_SERVER_GET_PRIVATE (server);
 	SoupMessage *msg;
 
 	soup_client_context_cleanup (client);
 
 	/* Listen for another request on this connection */
-	msg = g_object_new (SOUP_TYPE_MESSAGE, NULL);
-        soup_message_headers_set_encoding (msg->response_headers,
-                                           SOUP_ENCODING_CONTENT_LENGTH);
+	msg = g_object_new (SOUP_TYPE_MESSAGE,
+			    SOUP_MESSAGE_SERVER_SIDE, TRUE,
+			    NULL);
+	if (priv->server_header) {
+		soup_message_headers_append (msg->response_headers, "Server",
+					     priv->server_header);
+	}
 
 	g_signal_connect (msg, "got_headers", G_CALLBACK (got_headers), client);
 	g_signal_connect (msg, "got_body", G_CALLBACK (call_handler), client);
@@ -740,10 +892,9 @@ new_connection (SoupSocket *listner, SoupSocket *sock, gpointer user_data)
 {
 	SoupServer *server = user_data;
 	SoupServerPrivate *priv = SOUP_SERVER_GET_PRIVATE (server);
-	SoupClientContext *client = g_slice_new0 (SoupClientContext);
+	SoupClientContext *client;
 
-	client->server = server;
-	client->sock = g_object_ref (sock);
+	client = soup_client_context_new (server, g_object_ref (sock));
 	priv->client_socks = g_slist_prepend (priv->client_socks, sock);
 	g_signal_connect (sock, "disconnected",
 			  G_CALLBACK (socket_disconnected), server);
@@ -876,11 +1027,16 @@ soup_server_get_async_context (SoupServer *server)
 GType
 soup_client_context_get_type (void)
 {
-	static GType type = 0;
+	static volatile gsize type_volatile = 0;
 
-	if (type == 0)
-		type = g_pointer_type_register_static ("SoupClientContext");
-	return type;
+	if (g_once_init_enter (&type_volatile)) {
+		GType type = g_boxed_type_register_static (
+			g_intern_static_string ("SoupClientContext"),
+			(GBoxedCopyFunc) soup_client_context_ref,
+			(GBoxedFreeFunc) soup_client_context_unref);
+		g_once_init_leave (&type_volatile, type);
+	}
+	return type_volatile;
 }
 
 /**
@@ -1052,6 +1208,12 @@ soup_client_context_get_auth_user (SoupClientContext *client)
  * Adds a handler to @server for requests under @path. See the
  * documentation for #SoupServerCallback for information about
  * how callbacks should behave.
+ *
+ * If @path is %NULL or "/", then this will be the default handler for
+ * all requests that don't have a more specific handler. Note though
+ * that if you want to handle requests to the special "*" URI, you
+ * must explicitly register a handler for "*"; the default handler
+ * will not be used for that case.
  **/
 void
 soup_server_add_handler (SoupServer            *server,
@@ -1067,6 +1229,13 @@ soup_server_add_handler (SoupServer            *server,
 	g_return_if_fail (callback != NULL);
 	priv = SOUP_SERVER_GET_PRIVATE (server);
 
+	/* "" was never documented as meaning the same this as "/",
+	 * but it effectively was. We have to special case it now or
+	 * otherwise it would match "*" too.
+	 */
+	if (path && (!*path || !strcmp (path, "/")))
+		path = NULL;
+
 	hand = g_slice_new0 (SoupServerHandler);
 	hand->path       = g_strdup (path);
 	hand->callback   = callback;
@@ -1103,7 +1272,7 @@ soup_server_remove_handler (SoupServer *server, const char *path)
 	g_return_if_fail (SOUP_IS_SERVER (server));
 	priv = SOUP_SERVER_GET_PRIVATE (server);
 
-	if (!path) {
+	if (!path || !*path || !strcmp (path, "/")) {
 		if (priv->default_handler) {
 			unregister_handler (priv->default_handler);
 			free_handler (priv->default_handler);
@@ -1144,7 +1313,8 @@ soup_server_add_auth_domain (SoupServer *server, SoupAuthDomain *auth_domain)
 	g_return_if_fail (SOUP_IS_SERVER (server));
 	priv = SOUP_SERVER_GET_PRIVATE (server);
 
-	priv->auth_domains = g_slist_prepend (priv->auth_domains, auth_domain);
+	priv->auth_domains = g_slist_append (priv->auth_domains, auth_domain);
+	g_object_ref (auth_domain);
 }
 
 /**
@@ -1182,7 +1352,7 @@ soup_server_pause_message (SoupServer *server,
 	g_return_if_fail (SOUP_IS_SERVER (server));
 	g_return_if_fail (SOUP_IS_MESSAGE (msg));
 
-	soup_message_io_unpause (msg);
+	soup_message_io_pause (msg);
 }
 
 /**
