@@ -18,57 +18,100 @@ typedef struct {
 	const char *method;
 	const char *path;
 	guint status_code;
+	gboolean repeat;
 } TestRequest;
 
 static struct {
 	TestRequest requests[3];
+	guint final_status;
 } tests[] = {
-	/* A redirecty response to a GET should cause a redirect */
+	/* A redirecty response to a GET or HEAD should cause a redirect */
 
 	{ { { "GET", "/301", 301 },
 	    { "GET", "/", 200 },
-	    { NULL } } },
+	    { NULL } }, 200 },
 	{ { { "GET", "/302", 302 },
 	    { "GET", "/", 200 },
-	    { NULL } } },
+	    { NULL } }, 200 },
 	{ { { "GET", "/303", 303 },
 	    { "GET", "/", 200 },
-	    { NULL } } },
+	    { NULL } }, 200 },
 	{ { { "GET", "/307", 307 },
 	    { "GET", "/", 200 },
-	    { NULL } } },
+	    { NULL } }, 200 },
+	{ { { "HEAD", "/301", 301 },
+	    { "HEAD", "/", 200 },
+	    { NULL } }, 200 },
+	{ { { "HEAD", "/302", 302 },
+	    { "HEAD", "/", 200 },
+	    { NULL } }, 200 },
+	/* 303 is a nonsensical response to HEAD, but some sites do
+	 * it anyway. :-/
+	 */
+	{ { { "HEAD", "/303", 303 },
+	    { "HEAD", "/", 200 },
+	    { NULL } }, 200 },
+	{ { { "HEAD", "/307", 307 },
+	    { "HEAD", "/", 200 },
+	    { NULL } }, 200 },
 
-	/* A non-redirecty response to a GET should not */
+	/* A non-redirecty response to a GET or HEAD should not */
 
 	{ { { "GET", "/300", 300 },
-	    { NULL } } },
+	    { NULL } }, 300 },
 	{ { { "GET", "/304", 304 },
-	    { NULL } } },
+	    { NULL } }, 304 },
 	{ { { "GET", "/305", 305 },
-	    { NULL } } },
+	    { NULL } }, 305 },
 	{ { { "GET", "/306", 306 },
-	    { NULL } } },
+	    { NULL } }, 306 },
 	{ { { "GET", "/308", 308 },
-	    { NULL } } },
+	    { NULL } }, 308 },
+	{ { { "HEAD", "/300", 300 },
+	    { NULL } }, 300 },
+	{ { { "HEAD", "/304", 304 },
+	    { NULL } }, 304 },
+	{ { { "HEAD", "/305", 305 },
+	    { NULL } }, 305 },
+	{ { { "HEAD", "/306", 306 },
+	    { NULL } }, 306 },
+	{ { { "HEAD", "/308", 308 },
+	    { NULL } }, 308 },
 	
 	/* Test double-redirect */
 
 	{ { { "GET", "/301/302", 301 },
 	    { "GET", "/302", 302 },
-	    { "GET", "/", 200 } } },
+	    { "GET", "/", 200 } }, 200 },
+	{ { { "HEAD", "/301/302", 301 },
+	    { "HEAD", "/302", 302 },
+	    { "HEAD", "/", 200 } }, 200 },
 
-	/* POST should only automatically redirect on 302 and 303 */
+	/* POST should only automatically redirect on 301, 302 and 303 */
 
 	{ { { "POST", "/301", 301 },
-	    { NULL } } },
+	    { "GET", "/", 200 },
+	    { NULL } }, 200 },
 	{ { { "POST", "/302", 302 },
 	    { "GET", "/", 200 },
-	    { NULL } } },
+	    { NULL } }, 200 },
 	{ { { "POST", "/303", 303 },
 	    { "GET", "/", 200 },
-	    { NULL } } },
+	    { NULL } }, 200 },
 	{ { { "POST", "/307", 307 },
-	    { NULL } } }
+	    { NULL } }, 307 },
+
+	/* Test behavior with recoverably-bad Location header */
+	{ { { "GET", "/bad", 302 },
+	    { "GET", "/bad%20with%20spaces", 200 },
+	    { NULL } }, 200 },
+
+	/* Test behavior with irrecoverably-bad Location header */
+	{ { { "GET", "/bad-no-host", 302 },
+	    { NULL } }, SOUP_STATUS_MALFORMED },
+
+	{ { { "GET", "/bad-recursive", 302, TRUE },
+	    { NULL } }, SOUP_STATUS_TOO_MANY_REDIRECTS }
 };
 static const int n_tests = G_N_ELEMENTS (tests);
 
@@ -80,7 +123,8 @@ got_headers (SoupMessage *msg, gpointer user_data)
 
 	debug_printf (2, "    -> %d %s\n", msg->status_code,
 		      msg->reason_phrase);
-	location = soup_message_headers_get (msg->response_headers, "Location");
+	location = soup_message_headers_get_one (msg->response_headers,
+						 "Location");
 	if (location)
 		debug_printf (2, "       Location: %s\n", location);
 
@@ -102,7 +146,7 @@ restarted (SoupMessage *msg, gpointer user_data)
 
 	debug_printf (2, "    %s %s\n", msg->method, uri->path);
 
-	if ((*req)->method)
+	if ((*req)->method && !(*req)->repeat)
 		(*req)++;
 
 	if (!(*req)->method) {
@@ -150,6 +194,13 @@ do_test (SoupSession *session, SoupURI *base_uri, int n)
 			  G_CALLBACK (restarted), &req);
 
 	soup_session_send_message (session, msg);
+
+	if (msg->status_code != tests[n].final_status) {
+		debug_printf (1, "    - Expected final status of %d, got %d !\n",
+			      tests[n].final_status, msg->status_code);
+		errors++;
+	}
+
 	g_object_unref (msg);
 	debug_printf (2, "\n");
 }
@@ -164,15 +215,13 @@ do_redirect_tests (SoupURI *base_uri)
 	debug_printf (1, "Async session\n");
 	for (n = 0; n < n_tests; n++)
 		do_test (session, base_uri, n);
-	soup_session_abort (session);
-	g_object_unref (session);
+	soup_test_session_abort_unref (session);
 
 	session = soup_test_session_new (SOUP_TYPE_SESSION_SYNC, NULL);
 	debug_printf (1, "Sync session\n");
 	for (n = 0; n < n_tests; n++)
 		do_test (session, base_uri, n);
-	soup_session_abort (session);
-	g_object_unref (session);
+	soup_test_session_abort_unref (session);
 }
 
 static void
@@ -183,8 +232,32 @@ server_callback (SoupServer *server, SoupMessage *msg,
 	char *remainder;
 	guint status_code;
 
+	if (g_str_has_prefix (path, "/bad")) {
+		if (!strcmp (path, "/bad")) {
+			soup_message_set_status (msg, SOUP_STATUS_FOUND);
+			soup_message_headers_replace (msg->response_headers,
+						      "Location",
+						      "/bad with spaces");
+		} else if (!strcmp (path, "/bad-recursive")) {
+			soup_message_set_status (msg, SOUP_STATUS_FOUND);
+			soup_message_headers_replace (msg->response_headers,
+						      "Location",
+						      "/bad-recursive");
+		} else if (!strcmp (path, "/bad-no-host")) {
+			soup_message_set_status (msg, SOUP_STATUS_FOUND);
+			soup_message_headers_replace (msg->response_headers,
+						      "Location",
+						      "about:blank");
+		} else if (!strcmp (path, "/bad with spaces"))
+			soup_message_set_status (msg, SOUP_STATUS_OK);
+		else
+			soup_message_set_status (msg, SOUP_STATUS_NOT_FOUND);
+		return;
+	}
+
 	if (!strcmp (path, "/")) {
-		if (msg->method != SOUP_METHOD_GET) {
+		if (msg->method != SOUP_METHOD_GET &&
+		    msg->method != SOUP_METHOD_HEAD) {
 			soup_message_set_status (msg, SOUP_STATUS_METHOD_NOT_ALLOWED);
 			return;
 		}
@@ -204,9 +277,17 @@ server_callback (SoupServer *server, SoupMessage *msg,
 		}
 
 		soup_message_set_status (msg, SOUP_STATUS_OK);
-		soup_message_set_response (msg, "text/plain",
-					   SOUP_MEMORY_STATIC,
-					   "OK\r\n", 4);
+
+		/* FIXME: this is wrong, though it doesn't matter for
+		 * the purposes of this test, and to do the right
+		 * thing currently we'd have to set Content-Length by
+		 * hand.
+		 */
+		if (msg->method != SOUP_METHOD_HEAD) {
+			soup_message_set_response (msg, "text/plain",
+						   SOUP_MEMORY_STATIC,
+						   "OK\r\n", 4);
+		}
 		return;
 	}
 
@@ -257,7 +338,7 @@ main (int argc, char **argv)
 	loop = g_main_loop_new (NULL, TRUE);
 
 	if (run_tests) {
-		base_uri = soup_uri_new ("http://localhost");
+		base_uri = soup_uri_new ("http://127.0.0.1");
 		soup_uri_set_port (base_uri, port);
 		do_redirect_tests (base_uri);
 		soup_uri_free (base_uri);
