@@ -4,8 +4,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
-#include <netinet/in.h>
-#include <sys/socket.h>
 
 #include "libsoup/soup-address.h"
 #include "libsoup/soup-socket.h"
@@ -14,8 +12,14 @@
 #define BUFSIZE 1024
 #define DH_BITS 1024
 
-GMainLoop *loop;
-gnutls_dh_params_t dh_params;
+#ifndef G_OS_WIN32
+#define SOCKET_PRINT_ERROR(M) perror(M);
+#else
+#define SOCKET_PRINT_ERROR(M) g_error("%s: %d", M, WSAGetLastError());
+#endif
+
+static GMainLoop *loop;
+static gnutls_dh_params_t dh_params;
 
 /* SERVER */
 
@@ -64,8 +68,8 @@ server_write (gnutls_session_t session, char *buf, int bufsize)
 	}
 }
 
-const char *ssl_cert_file = SRCDIR "/test-cert.pem";
-const char *ssl_key_file = SRCDIR "/test-key.pem";
+static const char *ssl_cert_file = SRCDIR G_DIR_SEPARATOR_S "test-cert.pem";
+static const char *ssl_key_file = SRCDIR G_DIR_SEPARATOR_S "test-key.pem";
 
 static gpointer
 server_thread (gpointer user_data)
@@ -203,8 +207,6 @@ start_writing (gpointer user_data)
 	return FALSE;
 }
 
-int debug;
-
 static void
 debug_log (int level, const char *str)
 {
@@ -214,7 +216,7 @@ debug_log (int level, const char *str)
 int
 main (int argc, char **argv)
 {
-	int opt, listener, sin_len, port, i;
+	int opt, debug = 0, listener, sin_len, port, i;
 	struct sockaddr_in sin;
 	GThread *server;
 	char writebuf[BUFSIZE], readbuf[BUFSIZE];
@@ -225,8 +227,11 @@ main (int argc, char **argv)
 	SoupSocketIOStatus status;
 	GError *error = NULL;
 
-	g_type_init ();
 	g_thread_init (NULL);
+	g_type_init ();
+
+	/* On Windows, this will call WSAStartup() */
+	soup_socket_get_type ();
 
 	while ((opt = getopt (argc, argv, "c:d:k:")) != -1) {
 		switch (opt) {
@@ -255,21 +260,21 @@ main (int argc, char **argv)
 	/* Create server socket */
 	listener = socket (AF_INET, SOCK_STREAM, 0);
 	if (listener == -1) {
-		perror ("creating listening socket");
+		SOCKET_PRINT_ERROR ("creating listening socket");
 		exit (1);
 	}
 
 	memset (&sin, 0, sizeof (sin));
 	sin.sin_family = AF_INET;
-	sin.sin_addr.s_addr = INADDR_ANY;
+	sin.sin_addr.s_addr = ntohl (INADDR_LOOPBACK);
 
 	if (bind (listener, (struct sockaddr *) &sin, sizeof (sin))  == -1) {
-		perror ("binding listening socket");
+		SOCKET_PRINT_ERROR ("binding listening socket");
 		exit (1);
 	}
 
 	if (listen (listener, 1) == -1) {
-		perror ("listening on socket");
+		SOCKET_PRINT_ERROR ("listening on socket");
 		exit (1);
 	}
 
@@ -277,11 +282,7 @@ main (int argc, char **argv)
 	getsockname (listener, (struct sockaddr *)&sin, (void *)&sin_len);
 	port = ntohs (sin.sin_port);
 
-	/* Now spawn server thread */
-	server = g_thread_create (server_thread, GINT_TO_POINTER (listener),
-				  TRUE, NULL);
-
-	/* And create the client */
+	/* Create the client */
 	addr = soup_address_new ("127.0.0.1", port);
 	creds = soup_ssl_get_client_credentials (NULL);
 	sock = soup_socket_new (SOUP_SOCKET_REMOTE_ADDRESS, addr,
@@ -297,6 +298,10 @@ main (int argc, char **argv)
 
 	soup_socket_start_ssl (sock, NULL);
 
+	/* Now spawn server thread */
+	server = g_thread_create (server_thread, GINT_TO_POINTER (listener),
+				  TRUE, NULL);
+
 	/* Synchronous client test */
 	for (i = 0; i < BUFSIZE; i++)
 		writebuf[i] = i & 0xFF;
